﻿#include <nn.h>
#include <nn/snd.h>
#include <nn/dsp.h>
#include <nn/math.h>
#include "sndio.h"
#include "titlesample.h"


extern const u8 DSPAUDIO_BEGIN[];


class TestAllocator : public nn::fnd::IAllocator
{
private:
    nn::fnd::ExpHeap    expHeap;
    bool                isInitialized;
    NN_PADDING3;

public:
    TestAllocator(){isInitialized = false;};
    virtual ~TestAllocator(){};

    void InitExpHeap()
    {
        if(isInitialized == false)
        {
            expHeap.Initialize(
                nn::os::GetDeviceMemoryAddress(),
                8 * 1024 * 1024,
                nn::os::ALLOCATE_OPTION_LINEAR);
            isInitialized = true;
        }
    }

    virtual void* Allocate(size_t size, s32 alignment = 4)
    {
        if(isInitialized == true)
        {
            return expHeap.Allocate(size, alignment);
        }
        return NULL;
    }

    virtual void Free(void* p)
    {
        if(isInitialized == true)
        {
            expHeap.Free(p);
        }
    }
};


static const char* PAGE_NAME_TABLE[] =
{
    "CTR ",
    "SD01",
    "SD02",
    "SD03",
    "SD04",
    "SD05",
    "SD06",
    "SD07",
    "SD08",
    "SD09",
    "SD10"
};


void nnMain( void )
{
    NN_UNUSED_VAR( PAGE_NAME_TABLE );
    nn::fs::Initialize();
    TestAllocator allocator;

    nn::os::SetDeviceMemorySize( 8 * 1024 * 1024 );
    NN_UTIL_PANIC_IF_FAILED( nn::fs::MountSdmc() );

    // コンポーネントの初期化
    NN_UTIL_PANIC_IF_FAILED( nn::dsp::Initialize() );
    NN_UTIL_PANIC_IF_FAILED( nn::dsp::LoadComponent(DSPAUDIO_BEGIN, NN_DSPCOMPONENT_SIZE, 0xff, 0xff) );
    NN_UTIL_PANIC_IF_FAILED( nn::snd::Initialize() );

    NN_LOG( "start.\n" );

    sndio::Initialize();
    allocator.InitExpHeap();

    // ********** SNOTEで保存されている音声情報の取得 **********
    {
        for (int i = sndio::VOICE_PAGE_CTR_NAND; i < sndio::VOICE_PAGE_NUM; i++)
        {
            NN_LOG( " %s  ", PAGE_NAME_TABLE[i] );
        }
        NN_LOG( "\n" );
        for (int i = sndio::VOICE_PAGE_CTR_NAND; i < sndio::VOICE_PAGE_NUM; i++)
        {
            NN_LOG( "------ " );
        }
        NN_LOG( "\n" );

        for (int i = 0; i < 3; i++)
        {
            for (int j = sndio::VOICE_PAGE_CTR_NAND; j < sndio::VOICE_PAGE_NUM; j++)
            {
                for (int k = 0; k < 6; k++)
                {
                    NN_LOG( "%s", sndio::IsExistsEntry( static_cast<sndio::VoicePage>(j), i * 6 + k ) ? "O" : "x" );
                }

                NN_LOG( " " );
            }

            NN_LOG( "\n" );
        }

        for (int i = sndio::VOICE_PAGE_CTR_NAND; i < sndio::VOICE_PAGE_NUM; i++)
        {
            NN_LOG( "------ " );
        }
        NN_LOG( "\n" );
    }

    // ********** 音声情報を削除 **********
    {
        if (sndio::DeleteEntry( sndio::VOICE_PAGE_SD_10, 0 ))
        {
            // 情報の削除成功
            NN_LOG( "delete success.\n" );
        }
        else
        {
            // 情報の削除失敗
        }
    }

    sndio::VoiceEntry entry;

    // ********** 空いているところの音声情報の作成 **********
    {
        if (sndio::CreateEntry(
            &entry,
            sndio::VOICE_PAGE_SD_10,        // ページ
            0,                                // インデックス
            sndio::VOICE_COLOR_RED,            // ふきだしの色
            sndio::VOICE_FILL_50_LT,        // ふきだしの埋まり具合
            sndio::VOICE_FORM_NORMAL ))        // ふきだしの形
        {
            // 情報の作成成功
            NN_LOG( "%ls\n", entry.entryName );
        }
        else
        {
            // 情報の作成失敗
        }
    }

    // ********** 空いているところの音声情報にm4aとして保存 **********
    {
        NN_LOG( "voice write\n" );
        sndio::Result result;

        sndio::VoiceWriter writer( allocator );

        wchar_t titleName[ SNDIO_CONVERT_LENGTH ];
        nn::fnd::DateTime now = nn::fnd::DateTime::GetNow();

        ConvertDateTime( titleName, nn::cfg::CFG_REGION_JAPAN, nn::cfg::CFG_LANGUAGE_JAPANESE, now );
        writer.SetTitleName( titleName );
        writer.SetAlbumName( L"TEST" );
        writer.SetArtistName( L"TEST" );
        writer.SetCreateDate( now.GetParameters() );

        result = writer.StartStream( entry.entryName );
        if (result != sndio::ResultSuccess)
        {
            NN_LOG( "fail writer.StartStream %d\n", result );
        }

        s16* buf = reinterpret_cast<s16*>( allocator.Allocate( 8192, 32 ) );

        for (s32 i = 0; i < 4096; i++)
        {
            buf[i] = (32767.0f * nn::math::SinFIdx( 8.0f*i ));
        }
        nn::dsp::CTR::FlushDataCache( reinterpret_cast<uptr>(buf), 8192 );

        s32 writeCount = 0;
        for (u32 offset = 0; offset < 159744; offset += 4096)
        {
            result = writer.TryWrite( &writeCount, buf, 4096 );
            if (result != sndio::ResultSuccess)
            {
                NN_LOG( "fail writer.TryWrite %d\n", result );
                // エラー
                break;
            }
        }

        allocator.Free( buf );

        result = writer.EndStream();
        if (result != sndio::ResultSuccess)
        {
            NN_LOG( "fail writer.EndStream %d\n", result );
        }
    }

    // ********** デコードテスト **********
    {
        sndio::Result result;
        sndio::SoundReader reader;

        result = reader.Initialize( allocator, entry.entryName );

        if (result != sndio::ResultSuccess)
        {
            NN_LOG( "fail reader.Initialize %d\n", result );
        }

        NN_LOG( "----------------------------------------------\n" );
        NN_LOG( "format        : %s\n", reader.GetSoundFormat() == sndio::SOUND_FORMAT_AAC ? "AAC" : "MP3" );
        NN_LOG( "framesample   : %d\n", reader.GetFrameSamples() );
        NN_LOG( "framesize     : %d\n", reader.GetFrameBufferSize() );
        NN_LOG( "samplingrate  : %d\n", reader.GetSamplingRate() );
        NN_LOG( "channnel      : %d\n", reader.GetChannelCount() );
        NN_LOG( "totaltime     : %d\n", reader.GetTotalTime() );
        NN_LOG( "----------------------------------------------\n" );

        // オーディオフレームの読み込み
        {
            s16* pBufL = (s16*)allocator.Allocate( reader.GetFrameBufferSize(), 32 );
            s16* pBufR = (s16*)allocator.Allocate( reader.GetFrameBufferSize(), 32 );

            s32 readCount = 0;

            // シーク終了判定
            while (reader.GetSeekTime() < reader.GetTotalTime())
            {
                std::memset( pBufL, 0, reader.GetFrameBufferSize() );
                std::memset( pBufR, 0, reader.GetFrameBufferSize() );
                sndio::Result readResult = reader.TryRead( &readCount, pBufL, pBufR, reader.GetFrameSamples() );

                // TryReadの結果を無視して、次フレームを読み込む
                if (readResult != sndio::ResultSuccess)
                {
                     if (readResult == sndio::ResultInvalidDecoder)
                     {
                         // デコードエラーの場合は、readCountは0になっているので
                         // reader.GetFrameBufferSize()分時間を進めて再試行
                         continue;
                     }

                     // その他のエラーはエラーとして、処理を抜ける
                     break;
                }

                NN_LOG( "seek %d msec\n", reader.GetSeekTime() );
            }

            allocator.Free( pBufL );
            allocator.Free( pBufR );
        }

        reader.Finalize();
    }

    NN_LOG( "end.\n" );

    sndio::Finalize();

    nn::snd::Finalize();
    nn::dsp::UnloadComponent();
    nn::dsp::Finalize();
}
