/*---------------------------------------------------------------------------*
  Project:  sndio
  File:     sndio_SoundReader.cpp

  Copyright (C)2010 Nintendo Co., Ltd. All rights reserved

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: 19614 $
 *---------------------------------------------------------------------------*/

#include "sndio_SoundReader.h"
#include "detail/sndio_Mount.h"
#include "detail/sndio_Decoder.h"
#include "detail/sndio_MP3Seek.h"
#include "detail/sndio_CachedReader.h"
#include "detail/sndio_Util.h"

#if defined(USE_DSP)
  #include "detail/sndio_DecoderAACDsp.h"
  #if defined(USE_MP3)
    #include "detail/sndio_DecoderMP3Dsp.h"
  #endif
#else
#include "detail/sndio_DecoderAACArm.h"
  #if defined(USE_MP3)
    #include "detail/sndio_DecoderMP3Arm.h"
  #endif
#endif

#if defined(USE_DAT)
#include "detail/sndio_DecoderDSADPCM.h"
#endif

#if defined(USE_PCM)
#include "detail/sndio_DecoderPCM.h"
#endif

#include "detail/sndio_VoiceExtendTag.h"


namespace sndio {

using namespace detail;

namespace 
{
    //! gqTCY
    const u32 EXT_LENGTH = 4;

    //! wchar_t擾(obt@I[o[t[΍)
    static inline int _wcsnlen( const wchar_t* wstr, int maxLen )
    {
        register int __len;
        wchar_t* __s = const_cast<wchar_t*>(wstr);
        for (__len = 0; (!(*__s++ == L'\0')) && __len < maxLen; ++__len);
        return __len;
    }

    //! ϊ
    static inline wchar_t _toLower( wchar_t ch )
    {
        if (ch >= L'A' && ch <= L'Z')
        {
            ch = static_cast<wchar_t>(ch + L'a' - L'A');
        }
        return ch;
    }

    //! AACT|[gt@C
    static const wchar_t* supportAACExt[] =
    {
        L"m4a",
        L"3gp",
        L"mp4"
    };

    //! MP3T|[gt@C
    static const wchar_t* supportMP3Ext[] =
    {
        L"mp3"
    };

    //! DSTEht@C
    static const wchar_t* supportDSExt[] =
    {
        L"dat"
    };

    struct Kernel
    {
        nn::fnd::IAllocator*    m_allocator;
        IDecoder*               m_decoder;
        nn::fs::IInputStream*   m_stream;
        SoundFormat             m_format;
        NN_PADDING3;
        ExtendTag               m_tag;

        Kernel() :
        m_allocator( NULL ),
        m_decoder( NULL ),
        m_stream( NULL )
        {
            std::memset( &m_tag, 0, sizeof(m_tag) );
        }
    };
}

#define member( ins ) reinterpret_cast<Kernel*>(ins)


/*!
    @brief      TEh^Cv擾
*/
SoundFormat CheckFormat( const wchar_t* filePath )
{
    // t@CpXTCY`FbN
    int len = _wcsnlen(filePath, nn::fs::MAX_FILE_PATH_LENGTH);
    if (len < EXT_LENGTH && nn::fs::MAX_FILE_PATH_LENGTH < len)
    {
        // t@CpX邩Z
        return SOUND_FORMAT_NONE;
    }

    // gq𔲂o
    wchar_t ext[EXT_LENGTH];
    ext[0] = _toLower(filePath[ len - 3 ]);
    ext[1] = _toLower(filePath[ len - 2 ]);
    ext[2] = _toLower(filePath[ len - 1 ]);
    ext[3] = L'\0';

    // gqt@C^Cv𔻒f
    if (::wcscmp(supportAACExt[0], ext) == 0 ||
        ::wcscmp(supportAACExt[1], ext) == 0 ||
        ::wcscmp(supportAACExt[2], ext) == 0)
    {
        return SOUND_FORMAT_AAC;
    }
    else if (::wcscmp(supportMP3Ext[0], ext) == 0)
    {
        return SOUND_FORMAT_MP3;
    }
    else if (::wcscmp(supportDSExt[0], ext) == 0)
    {
        return SOUND_FORMAT_DAT;
    }

    // ΉtH[}bg
    return SOUND_FORMAT_NONE;
}


/*!
    @brief      \z

                RXgN^łB<br/>
*/
SoundReader::SoundReader() :
m_kernel( NULL ),
m_stream( NULL )
{
    std::memset( m_filePath, 0, sizeof(m_filePath) );
}


/*!
    @brief      

                fXgN^łB<br/>
*/
SoundReader::~SoundReader()
{
    Finalize();
}


/*!
    @brief      s܂B

                AP[^̓foCX̊mۂłKv܂B<br/>
                ADTS-AAC(.aac)t@C͎wł܂B<br/>

    @param[in]  allocator   AP[^(foCX)
    @param[in]  filePath    t@CpX(M4A,MP4,3GP,MP3̃t@C)

    @retval     ResultSuccess               
    @retval     ResultError                 sȃG[
    @retval     ResultFileNotFound          t@C݂Ȃ
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultOutOfResource         \[Xs
    @retval     ResultNotFinalize           JĂȂ
    @retval     ResultInvalidFile           T|[g̃t@C
    @retval     ResultInvalidDecoder        fR[_[G[
*/
Result SoundReader::Initialize(
    nn::fnd::IAllocator& allocator,
    const wchar_t* filePath )
{
    CachedReader* pReader = 
        new (allocator.Allocate( sizeof(CachedReader), 32 )) CachedReader();

    // zpXȂϊ
    std::wcsncpy( m_filePath, filePath, nn::fs::MAX_FILE_PATH_LENGTH );
    if (IsVirtual( m_filePath ))
    {
        ToReal( m_filePath );
    }

	if (pReader->Open( allocator, m_filePath ))
    {
        m_stream = pReader;
        return Initialize( allocator, *m_stream, CheckFormat( filePath ) );
    }

    pReader->Close();
    SafeFree( &allocator, pReader );

    return ResultInvalidFile;
}


/*!
    @brief      Xg[珉s܂B

                AP[^̓foCX̊mۂłKv܂B<br/>
                ADTS-AAC(.aac)t@C͎wł܂B<br/>

    @param[in]  allocator   AP[^(foCX)
    @param[in]  stream      Xg[
    @param[in]  format      TEhtH[}bg

    @retval     ResultSuccess               
    @retval     ResultError                 sȃG[
    @retval     ResultFileNotFound          t@C݂Ȃ
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultOutOfResource         \[Xs
    @retval     ResultNotFinalize           JĂȂ
    @retval     ResultInvalidFile           T|[g̃t@C
    @retval     ResultInvalidParameter      p[^s
    @retval     ResultInvalidDecoder        fR[_[G[
*/
Result SoundReader::Initialize(
    nn::fnd::IAllocator& allocator,
    nn::fs::IInputStream& stream,
    SoundFormat format )
{
    if (m_kernel)
    {
        return ResultNotFinalize;
    }

    m_kernel = allocator.Allocate( sizeof(Kernel), 32 );
    m_kernel = new(m_kernel) Kernel();

    if (m_kernel == NULL)
    {
        return ResultOutOfResource;
    }

    member(m_kernel)->m_allocator = &allocator;

    switch (format)
    {
#if defined(USE_DSP)
    case SOUND_FORMAT_AAC:
        {
            member(m_kernel)->m_format = SOUND_FORMAT_AAC;

            // AACfR[_[
            void* pDecoder = allocator.Allocate( sizeof(DecoderAACDsp), 32 );
            member(m_kernel)->m_decoder = new (pDecoder) DecoderAACDsp();
        }
        break;

#if defined(USE_MP3)
    case SOUND_FORMAT_MP3:
        {
            member(m_kernel)->m_format = SOUND_FORMAT_MP3;

            // MP3fR[_[
            void* pDecoder = allocator.Allocate( sizeof(DecoderMP3Dsp), 32 );
            member(m_kernel)->m_decoder = new (pDecoder) DecoderMP3Dsp();
        }
        break;
#endif

#else
    case SOUND_FORMAT_AAC:
        {
            member(m_kernel)->m_format = SOUND_FORMAT_AAC;

            // AACfR[_[
            void* pDecoder = allocator.Allocate( sizeof(DecoderAACArm), 32 );
            member(m_kernel)->m_decoder = new (pDecoder) DecoderAACArm();
        }
        break;

#if defined(USE_MP3)
    case SOUND_FORMAT_MP3:
        {
            member(m_kernel)->m_format = SOUND_FORMAT_MP3;

            // MP3fR[_[
            void* pDecoder = allocator.Allocate( sizeof(DecoderMP3Arm), 32 );
            member(m_kernel)->m_decoder = new (pDecoder) DecoderMP3Arm();
        }
        break;
#endif
#endif

#if defined(USE_DAT)
    case SOUND_FORMAT_DAT:
        {
            member(m_kernel)->m_format = SOUND_FORMAT_DAT;

            // DATfR[_[
            void* pDecoder = allocator.Allocate( sizeof(DecoderDSADPCM), 32 );
            member(m_kernel)->m_decoder = new (pDecoder) DecoderDSADPCM();
        }
        break;
#endif

#if defined(USE_PCM)
    case SOUND_FORMAT_PCM:
        {
            member(m_kernel)->m_format = SOUND_FORMAT_PCM;

            // PCMfR[_[
            void* pDecoder = allocator.Allocate( sizeof(DecoderPCM), 32 );
            member(m_kernel)->m_decoder = new (pDecoder) DecoderPCM();
        }
        break;
#endif

    default:
        return ResultInvalidFile;
    }

    if (member(m_kernel)->m_decoder == NULL)
    {
        Finalize();
        return ResultOutOfResource;
    }

    std::memset( &member(m_kernel)->m_tag, 0x00, sizeof(ExtendTag) );

    Result result = member(m_kernel)->m_decoder->Initialize(
        &allocator,
        &stream,
        reinterpret_cast<ExtendTag*>(&member(m_kernel)->m_tag)
    );

    if (result == ResultSuccess)
    {
        return ResultSuccess;
    }

    // Ɏs
    Finalize();
    return result;
}


/*!
    @brief      j܂B

                fXgN^̒łĂ΂܂B<br/>
                IɌĂł肠܂B<br/>
*/
void SoundReader::Finalize()
{
    if (m_stream)
    {
        nn::fnd::IAllocator* allocator = member(m_kernel)->m_allocator;
        NN_NULL_TASSERT_( allocator );

        if (allocator)
        {
            m_stream->~IInputStream();
            SafeFree( allocator, m_stream );
        }
        m_stream = NULL;
    }

    if (m_kernel)
    {
        nn::fnd::IAllocator* allocator = member(m_kernel)->m_allocator;
        NN_NULL_TASSERT_( allocator );

        if (allocator && member(m_kernel)->m_decoder)
        {
            member(m_kernel)->m_decoder->~IDecoder();
            SafeFree( allocator, member(m_kernel)->m_decoder );
        }

        SafeFree( allocator, m_kernel );
        m_kernel = NULL;
    }
}


/*!
    @brief      ^Cg擾܂B
    @return     ^Cgւ̃|C^
*/
const wchar_t* SoundReader::GetTitleName() const
{
    if (m_kernel)
    {
        if (member(m_kernel)->m_decoder)
        {
            return member(m_kernel)->m_decoder->GetTitleName();
        }
    }

    return NULL;
}


/*!
    @brief      Ao擾܂
    @return     Aoւ̃|C^
*/
const wchar_t* SoundReader::GetAlbumName() const
{
    if (m_kernel)
    {
        if (member(m_kernel)->m_decoder)
        {
            return member(m_kernel)->m_decoder->GetAlbumName();
        }
    }

    return NULL;
}


/*!
    @brief      A[eBXg擾܂
    @return     A[eBXgւ̃|C^
*/
const wchar_t* SoundReader::GetArtistName() const
{
    if (m_kernel)
    {
        if (member(m_kernel)->m_decoder)
        {
            return member(m_kernel)->m_decoder->GetArtistName();
        }
    }

    return NULL;
}


/*!
    @brief      g^O񂪗L`FbN܂
    @return     Lȏꍇtrue
*/
bool SoundReader::IsExistExtendTag() const
{
    if (m_kernel &&
        member(m_kernel)->m_tag.code[0] == '3' &&
        member(m_kernel)->m_tag.code[1] == 'D' &&
        member(m_kernel)->m_tag.code[2] == 'S' &&
        member(m_kernel)->m_tag.code[3] == '1')
    {
        return true;
    }

    return false;
}


/*!
    @brief      ۑ\tgEFAID擾܂

    @param[out] \tgEFAID擾
    @return     擾łꍇtrue
*/
bool SoundReader::GetSFTW( u8 sftw[5] ) const
{
    if (IsExistExtendTag())
    {
        std::memcpy( sftw, member(m_kernel)->m_tag.sftw, 5 );
        return true;
    }

    return false;
}


/*!
    @brief      ^{ID擾܂

    @param[out] ^{ID擾
    @return     擾łꍇtrue
*/
bool SoundReader::GetRCID( u8 rcid[8] ) const
{
    if (IsExistExtendTag())
    {
        std::memcpy( rcid, member(m_kernel)->m_tag.rcid, 8 );
        return true;
    }

    return false;
}


/*!
    @brief      CTRAvŕҏW{ID擾܂

    @param[out] CTRAvŕҏW{ID擾
    @return     擾łꍇtrue
*/
bool SoundReader::GetEDID( u8 edid[8] ) const
{
    if (IsExistExtendTag())
    {
        std::memcpy( edid, member(m_kernel)->m_tag.edid, 8 );
        return true;
    }

    return false;
}


/*!
    @brief      Abv[hĂ邩擾܂
    @return     0: 1:C^[lbgAbv[h
*/
u8 SoundReader::GetPermitUpload() const
{
    if (m_kernel)
    {
        return member(m_kernel)->m_tag.permitUpload;
    }

    return 0;
}


/*!
    @brief      ό̉zzĂ邩擾܂
    @return     0: 1:zz
*/
u8 SoundReader::GetPermitEditDistribute() const
{
    if (m_kernel)
    {
        return member(m_kernel)->m_tag.permitEditDistribute;
    }

    return 0;
}


/*!
    @brief      \擾܂
    @return     0:s 1:
*/
u8 SoundReader::GetPermitFirstDistribute() const
{
    if (m_kernel)
    {
        return member(m_kernel)->m_tag.permitFirstDistribute;
    }

    return 0;
}


/*!
    @brief      񎟔zz\擾܂
    @return     0:񎟔zzs 1:񎟔zz
*/
u8 SoundReader::GetPermitSecondDistribute() const
{
    if (m_kernel)
    {
        return member(m_kernel)->m_tag.permitSecondDistribute;
    }

    return 0;
}


/*!
    @brief      TEhtH[}bg擾܂B
    @return     TEhtH[}bg
*/
SoundFormat SoundReader::GetSoundFormat() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_format;
    }

    return SOUND_FORMAT_NONE;
}


/*!
    @brief      Pt[̃Tv擾܂B
    @return     Tv
*/
u32 SoundReader::GetFrameSamples() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->GetFrameSamples();
    }

    return 0;
}


/*!
    @brief      Pt[̃obt@TCY(Byte)擾܂B
    @return     obt@TCY(Byte)
*/
u32 SoundReader::GetFrameBufferSize() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->GetFrameBufferSize();
    }

    return 0;
}


/*!
    @brief      TvO[g擾܂B
    @return     TvO[g
*/
s32 SoundReader::GetSamplingRate() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->GetSamplingRate();
    }

    return 0;
}


/*!
    @brief      `l擾܂B
    @return     `l
*/
s32 SoundReader::GetChannelCount() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->GetChannelCount();
    }

    return 0;
}


/*!
    @brief      ĐԂ擾܂(msecP)
    @return     Đ(msecP)
*/
s32 SoundReader::GetTotalTime() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->GetTotalTime();
    }

    return 0;
}


/*!
    @brief      V[NԂ擾܂(msecP)
    @return     V[N(msecP)
*/
s32 SoundReader::GetSeekTime() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->GetSeekTime();
    }

    return 0;
}


/*!
    @brief      V[N\
    @return     true:V[N\, false:V[Ns
*/
bool SoundReader::CanSeek() const
{
    if (m_kernel)
    {
        NN_NULL_TASSERT_( member(m_kernel)->m_decoder );
        return member(m_kernel)->m_decoder->CanSeek();
    }

    return false;
}


/*!
    @brief      (msecP)ŃV[Nʒuݒ肵܂
    @param[in]  V[N(msecP)
*/
void SoundReader::SetSeekTime( s32 seekTime )
{
    if (m_kernel)
    {
        member( m_kernel )->m_decoder->SetSeekTime( seekTime );
    }
}


/*!
    @brief      tĐs(AAC-DSP̂)
    @param[in]  true tĐw, false Đw(ftHg)
*/
void SoundReader::SetReverse( bool reverse )
{
    if (m_kernel)
    {
        member( m_kernel )->m_decoder->SetReverse( reverse );
    }
}


/*!
    @brief      TEh̔g`f[^(PCM16)擾܂B

                fR[hŝŎsɂ͎Ԃ܂B<br/>
                󂯎obt@32oCgACg̃foCX
                Kv܂B<br/>
                m̏ꍇ́ApBufR𖳉(0)ɐݒ肵܂B<br/>
                AACłΎ擾Tv𒴂Ȃ1024̔{Ŏ擾܂B<br/>
                MP3łΎ擾Tv𒴂Ȃ576,1152̔{Ŏ擾܂B

    @param[out] readSamples ǂ݂ƂTv
    @param[out] pBufL       `l󂯎obt@
    @param[out] pBufR       E`l󂯎obt@
    @param[in]  samples     擾Tv

    @retval     ResultSuccess               ǂݎ萬
    @retval     ResultError                 sȃG[
    @retval     ResultMediaNotFound         fBAĂȂ
    @retval     ResultMediaInvalid          fBAs
    @retval     ResultAlreadyExists         t@Cɑ
    @retval     ResultNotEnoughSpace        eʕsG[
    @retval     ResultBadFormat             stH[}bg
    @retval     ResultArchiveInvalidated    A[JCus
    @retval     ResultOperationDenied       ANZX
    @retval     ResultWriteProtected        ݋֎~G[
    @retval     ResultMediaAccessError      fBAANZXG[
    @retval     ResultNotFormatted          tH[}bg
    @retval     ResultOutOfResource         \[Xs
    @retval     ResultInvalidFile           T|[g̃t@C
    @retval     ResultInvalidParameter      p[^s
    @retval     ResultInvalidDecoder        fR[_[G[
*/
Result SoundReader::TryRead( s32* readCount, s16* pBufL, s16* pBufR, s32 samples )
{
    if (m_kernel)
    {
        return member( m_kernel )->m_decoder->TryRead( readCount, pBufL, pBufR, samples );
    }

    return ResultInvalidFile;
}


} // namespace sndio
